import os
from typing import Dict
import logging
import torch
import tiinfer
import base64
import mosec

from diffusers import StableDiffusionControlNetPipeline

### 如果使用加速，需要导入如下代码
import tiacc_inference
TIACC_PATH = "model/tiacc"
### 如果使用加速，需要导入如上代码

# 未使用加速时模型路径
# MODEL_PATH = "."

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    "%(asctime)s - %(process)d - %(levelname)s - %(filename)s:%(lineno)s - %(message)s"
)
sh = logging.StreamHandler()
sh.setFormatter(formatter)
logger.addHandler(sh)


class Infer(mosec.Worker):
    def __init__(self) -> None:
        super().__init__()
        # 获取当前所在目录，需要据此加载模型
        self.root_path = tiinfer.get_ti_model_dir()

        ### 原始模型加载,若使用加速模型，需要注释掉这段
        #model_id = os.path.join(self.root_path, MODEL_PATH)
        #pipe = StableDiffusionControlNetPipeline.from_pretrained(model_id, torch_dtype=torch.float16)
        #self.model = pipe
        ### 原始模型加载,若使用加速模型，需要注释掉这段

        ### 加速模型,需要导入这段
        model_id = os.path.join(self.root_path, TIACC_PATH)
        pipe = tiacc_inference.load(model_id)[0]
        self.model = pipe.to("cuda")
        ## 预热一下
        ### 加速模型,需要导入这段


        canny_image_path = 'https://tione-prod-open-1256580188.cos.ap-guangzhou.myqcloud.com/ti-cloud/ti-ems/images/input_image_vermeer.png'
        canny_image = gen_controlnet_canny_img(canny_image_path)
        # 预热一下模型
        for i in range(2):
            img = self.model("a photo of an astronaut riding a horse on mars", canny_image).images[0]
        torch.cuda.synchronize()

        try:
            self.uploader = CosUploader()
        except Exception as e:
            self.uploader = None
            logging.info(f'no cos uploader found: {e}')

    def forward(self, req: Dict) -> Dict:
        prompt = req["prompt"]
        img_path = req["image"]
        canny_image = gen_controlnet_canny_img(img_path)
        image = self.model(prompt, canny_image).images[0]
        url = None
        if self.uploader is not None:
            url = self.uploader.upload_image(image)
        if url is None or len(url) == 0:
            bs = bytes_from_image(image)
            return {
                "image": base64.b64encode(bs).decode('ascii')
            }
        return {"image": url}

tiinfer.append_worker(Infer, 1)

def bytes_from_image(image):
    import io
    buffer = io.BytesIO()
    image.save(buffer, format='PNG')
    return buffer.getvalue()


class CosUploader:
    def __init__(self):
        if os.getenv('COS_UPLOAD_ENABLE') is None:
            raise ValueError("environment COS_UPLOAD_ENABLE is not set")
        if os.getenv('COS_SECRET_ID') is None:
            raise ValueError("environment COS_SECRET_ID is not set")
        if os.getenv('COS_SECRET_KEY') is None:
            raise ValueError("environment COS_SECRET_KEY is not set")
        if os.getenv('COS_BUCKET') is None:
            raise ValueError("environment COS_BUCKET is not set")
        if os.getenv('COS_REGION') is None:
            raise ValueError("environment COS_REGION is not set")

        # 1. 设置用户属性, 包括 secret_id, secret_key, region 等。Appid 已在 CosConfig 中移除，请在参数 Bucket 中带上 Appid。Bucket 由 BucketName-Appid 组成
        # 用户的 SecretId，建议使用子账号密钥，授权遵循最小权限指引，降低使用风险。子账号密钥获取可参见 https://cloud.tencent.com/document/product/598/37140
        secret_id = os.getenv("COS_SECRET_ID")
        # 用户的 SecretKey，建议使用子账号密钥，授权遵循最小权限指引，降低使用风险。子账号密钥获取可参见 https://cloud.tencent.com/document/product/598/37140
        secret_key = os.getenv("COS_SECRET_KEY")
        # 替换为用户的 region，已创建桶归属的 region 可以在控制台查看，https://console.cloud.tencent.com/cos5/bucket
        # COS 支持的所有 region 列表参见 https://cloud.tencent.com/document/product/436/6224
        region = os.getenv("COS_REGION")
        # 如果使用永久密钥不需要填入 token，如果使用临时密钥需要填入，临时密钥生成和使用指引参见 https://cloud.tencent.com/document/product/436/14048
        token = None               
        # 指定使用 http/https 协议来访问 COS，默认为 https，可不填
        scheme = 'https'
        self.bucket = os.getenv('COS_BUCKET')
        self.prefix = os.getenv('COS_FILE_PREFIX', '')

        from qcloud_cos import CosConfig
        from qcloud_cos import CosS3Client
        config = CosConfig(Region=region, SecretId=secret_id, SecretKey=secret_key, Token=token, Scheme=scheme)
        self.client = CosS3Client(config)
    
    def upload_image(self, image, key: str=None) -> str:
        if key is None:
            import uuid
            key = str(uuid.uuid4()) + '.png'
        if len(self.prefix) > 0:
            key = os.path.join(self.prefix, key)
        from qcloud_cos.cos_exception import CosClientError, CosServiceError

        # 使用高级接口断点续传，失败重试时不会上传已成功的分块(这里重试10次)
        for i in range(0, 10):
            try:
                response = self.client.put_object(
                    Bucket=self.bucket,
                    Body=bytes_from_image(image),
                    Key=key)
                url = self.client.get_object_url(
                    Bucket=self.bucket,
                    Key=key,
                )
                return url
            except (CosClientError, CosServiceError) as e:
                logging.error(f"failed to upload image: {e}")
        return ""


# 从原始图像获取canny图，参考https://huggingface.co/docs/diffusers/main/en/api/pipelines/stable_diffusion/controlnet
def gen_controlnet_canny_img(canny_image_path):
    import PIL.Image
    import PIL.ImageOps
    from packaging import version

    if version.parse(version.parse(PIL.__version__).base_version) >= version.parse("9.1.0"):
        PIL_INTERPOLATION = {
            "linear": PIL.Image.Resampling.BILINEAR,
            "bilinear": PIL.Image.Resampling.BILINEAR,
            "bicubic": PIL.Image.Resampling.BICUBIC,
            "lanczos": PIL.Image.Resampling.LANCZOS,
            "nearest": PIL.Image.Resampling.NEAREST,
        }
    else:
        PIL_INTERPOLATION = {
            "linear": PIL.Image.LINEAR,
            "bilinear": PIL.Image.BILINEAR,
            "bicubic": PIL.Image.BICUBIC,
            "lanczos": PIL.Image.LANCZOS,
            "nearest": PIL.Image.NEAREST,
        }
    from diffusers.utils import load_image
    image = load_image(
        canny_image_path
    )
    import cv2
    from PIL import Image
    import numpy as np

    image = np.array(image)

    low_threshold = 100
    high_threshold = 200

    image = cv2.Canny(image, low_threshold, high_threshold)
    image = image[:, :, None]
    image = np.concatenate([image, image, image], axis=2)
    canny_image = Image.fromarray(image)

    in_image = [canny_image]
    height = canny_image.height
    height = (height // 8) * 8
    width = canny_image.width
    width = (width // 8) * 8
    in_image = [
        np.array(i.resize((width, height), resample=PIL_INTERPOLATION["lanczos"]))[None, :] for i in in_image
    ]
    in_image = np.concatenate(in_image, axis=0)
    in_image = np.array(in_image).astype(np.float32) / 255.0
    in_image = in_image.transpose(0, 3, 1, 2)
    in_image = torch.from_numpy(in_image)
    return in_image